"""
Usage:
    local_runner.py --raw-data=<files> --streaming-dtc=<file> [--window-dtc=<file>] [--output-file=<file>]
    local_runner.py (-h | --help)
"""
import csv
import json
from collections import defaultdict
from datetime import datetime
from typing import List, Optional, Any, Dict, Tuple

from blurr.core.record import Record
from blurr.core.syntax.schema_validator import validate
from blurr.runner.data_processor import DataProcessor, SimpleJsonDataProcessor
from blurr.runner.runner import Runner, BlurrJSONEncoder


class LocalRunner(Runner):
    def __init__(self, stream_dtc_file: str, window_dtc_file: Optional[str] = None):
        super().__init__(stream_dtc_file, window_dtc_file)

        self._block_data = {}
        self._window_data = defaultdict(list)

    def _validate_dtc_syntax(self) -> None:
        validate(self._stream_dtc)
        if self._window_dtc is not None:
            validate(self._window_dtc)

    def _execute_for_all_identities(
            self, identity_records: Dict[str, List[Tuple[datetime, Record]]]) -> None:
        for per_identity_records in identity_records.items():
            data = self.execute_per_identity_records(per_identity_records)
            if self._window_dtc:
                self._window_data.update(data)
            else:
                self._block_data.update(data)

    def get_identity_records_from_json_files(
            self,
            local_json_files: List[str],
            data_processor: DataProcessor = SimpleJsonDataProcessor()
    ) -> Dict[str, List[Tuple[datetime, Record]]]:
        identity_records = defaultdict(list)
        for file in local_json_files:
            with open(file) as f:
                for identity, record_with_datetime in self.get_per_identity_records(f, data_processor):
                    identity_records[identity].append(record_with_datetime)
        return identity_records

    def execute(self, identity_records: Dict[str, List[Tuple[datetime, Record]]]) -> Any:
        self._execute_for_all_identities(identity_records)
        return self._window_data if self._window_dtc else self._block_data

    def print_output(self, data) -> None:
        for row in data.items():
            print(json.dumps(row, cls=BlurrJSONEncoder))

    def write_output_file(self, output_file: str, data):
        if not self._window_dtc:
            with open(output_file, 'w') as file:
                for row in data.items():
                    file.write(json.dumps(row, cls=BlurrJSONEncoder))
                    file.write('\n')
        else:
            header = []
            for data_rows in data.values():
                for data_row in data_rows:
                    header = list(data_row.keys())
                    break
            header.sort()
            with open(output_file, 'w') as csv_file:
                writer = csv.DictWriter(csv_file, header)
                writer.writeheader()
                for data_rows in data.values():
                    writer.writerows(data_rows)



from typing import Any, Dict, List, Tuple

import boto3
from boto3.dynamodb.conditions import Key as DynamoKey
from dateutil import parser

from blurr.core.schema_loader import SchemaLoader
from blurr.core.store import Store, Key

ATTRIBUTE_TABLE = 'Table'
ATTRIBUTE_READ_CAPACITY_UNITS = 'ReadCapacityUnits'
ATTRIBUTE_WRITE_CAPACITY_UNITS = 'WriteCapacityUnits'


class DynamoStore(Store):
    """
    In-memory store implementation
    """

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        super().__init__(fully_qualified_name, schema_loader)
        spec = schema_loader.get_schema_spec(fully_qualified_name)
        self.table_name = spec[ATTRIBUTE_TABLE]
        self.rcu = spec.get(ATTRIBUTE_READ_CAPACITY_UNITS, 5)
        self.wcu = spec.get(ATTRIBUTE_WRITE_CAPACITY_UNITS, 5)

        self.dynamodb_resource = boto3.resource('dynamodb')

        self.table = self.dynamodb_resource.Table(self.table_name)

        # Test that the table exists.  Create a new one otherwise
        try:
            self.table.creation_date_time
        except self.dynamodb_resource.meta.client.exceptions.ResourceNotFoundException:
            self.table = self.dynamodb_resource.create_table(
                TableName=self.table_name,
                KeySchema=[
                    {
                        'AttributeName': 'partition_key',
                        'KeyType': 'HASH'
                    },
                    {
                        'AttributeName': 'range_key',
                        'KeyType': 'RANGE'
                    },
                ],
                AttributeDefinitions=[{
                    'AttributeName': 'partition_key',
                    'AttributeType': 'S'
                }, {
                    'AttributeName': 'range_key',
                    'AttributeType': 'S'
                }],
                ProvisionedThroughput={
                    'ReadCapacityUnits': self.rcu,
                    'WriteCapacityUnits': self.wcu
                })
            # Wait until the table creation is complete
            self.table.meta.client.get_waiter('table_exists').wait(
                TableName=self.table_name, WaiterConfig={'Delay': 5})

    @staticmethod
    def dimensions(key: Key):
        return key.group + (key.PARTITION + key.timestamp.isoformat() if key.timestamp else '')

    @staticmethod
    def clean_for_get(item: Dict[str, Any]) -> Dict[str, Any]:
        item.pop('partition_key', None)
        item.pop('range_key', None)
        return item

    @staticmethod
    def clean_item_for_save(item: Dict[str, Any]) -> Dict[str, Any]:
        return {k: v for k, v in item.items() if v}

    def prepare_record(self, record: Dict[str, Any]) -> Tuple[Key, Any]:
        dimensions = record['range_key'].split(Key.PARTITION)
        key = Key(record['partition_key'], dimensions[0], None
                  if len(dimensions) == 1 else parser.parse(dimensions[1]))
        return key, self.clean_for_get(record)

    def get(self, key: Key) -> Any:
        item = self.table.get_item(Key={
            'partition_key': key.identity,
            'range_key': self.dimensions(key)
        }).get('Item', None)

        if not item:
            return None

        return self.clean_for_get(item)

    def get_range(self, start: Key, end: Key = None, count: int = 0) -> List[Tuple[Key, Any]]:

        if end and count:
            raise ValueError('Only one of `end` or `count` can be set')

        if end is not None and end < start:
            start, end = end, start

        dimension_key_condition = DynamoKey('range_key')

        if end:
            dimension_key_condition = dimension_key_condition.between(
                self.dimensions(start), self.dimensions(end))
        else:
            dimension_key_condition = dimension_key_condition.gt(
                self.dimensions(start)) if count > 0 else dimension_key_condition.lt(
                    self.dimensions(start))

        response = self.table.query(
            Limit=abs(count) if count else 1000,
            KeyConditionExpression=DynamoKey('partition_key').eq(start.identity) &
            dimension_key_condition,
            ScanIndexForward=count >= 0,
        )

        records = [self.prepare_record(item)
                   for item in response['Items']] if 'Items' in response else []

        if not records:
            return records

        # Ignore the starting record because `between` includes the records that match the boundary condition
        if records[0][0] == start or records[0][0] == end:
            del records[0]

        if records[-1][0] == start or records[-1][0] == end:
            del records[-1]

        return records

    def get_all(self, identity: str) -> Dict[Key, Any]:
        response = self.table.query(KeyConditionExpression=DynamoKey('partition_key').eq(identity))
        return dict([self.prepare_record(item)
                     for item in response['Items']] if 'Items' in response else [])

    def save(self, key: Key, item: Any) -> None:
        item['partition_key'] = key.identity
        item['range_key'] = self.dimensions(key)
        self.table.put_item(Item=self.clean_item_for_save(item))

    def delete(self, key: Key) -> None:
        pass

    def finalize(self) -> None:
        pass



import json
from typing import List, Optional

from blurr.runner.data_processor import DataProcessor, SimpleJsonDataProcessor, \
    SimpleDictionaryDataProcessor
from blurr.runner.runner import Runner, BlurrJSONEncoder

_spark_import_err = None
try:
    from pyspark import RDD, SparkContext
    from pyspark.sql import SparkSession
except ImportError as err:
    # Ignore import error because the CLI can be used even if spark is not
    # installed.
    _spark_import_err = err

# Setting these at module level as they cannot be part of the spark runner object because they
# cannot be serialized
_module_spark_session: 'SparkSession' = None


def get_spark_session(spark_session: Optional['SparkSession']) -> 'SparkSession':
    if spark_session:
        return spark_session

    global _module_spark_session
    if _module_spark_session:
        return _module_spark_session

    _module_spark_session = SparkSession \
        .builder \
        .appName("BlurrSparkRunner") \
        .getOrCreate()
    return _module_spark_session


class SparkRunner(Runner):
    def __init__(self, stream_dtc_file: str, window_dtc_file: Optional[str] = None):
        if _spark_import_err:
            raise _spark_import_err
        super().__init__(stream_dtc_file, window_dtc_file)

    def execute(self, identity_records: 'RDD'):
        return identity_records.flatMap(lambda x: self.execute_per_identity_records(x))

    def get_record_rdd_from_json_files(self,
                                       json_files: List[str],
                                       data_processor: DataProcessor = SimpleJsonDataProcessor(),
                                       spark_session: Optional['SparkSession'] = None) -> 'RDD':
        spark_context = get_spark_session(spark_session).sparkContext
        raw_records: 'RDD' = spark_context.union([spark_context.textFile(file) for file in json_files])
        return raw_records.mapPartitions(
            lambda x: self.get_per_identity_records(x, data_processor)).groupByKey().mapValues(list)

    def get_record_rdd_from_rdd(self,
                                rdd: 'RDD',
                                data_processor: DataProcessor = SimpleDictionaryDataProcessor()
                                ) -> 'RDD':
        return rdd.mapPartitions(
            lambda x: self.get_per_identity_records(x, data_processor)).groupByKey().mapValues(list)

    def write_output_file(self,
                          path: str,
                          per_identity_data: 'RDD',
                          spark_session: Optional['SparkSession'] = None) -> None:
        _spark_session_ = get_spark_session(spark_session)
        if not self._window_dtc:
            per_identity_data.map(lambda x: json.dumps(x, cls=BlurrJSONEncoder)).saveAsTextFile(
                path)
        else:
            # Convert to a DataFrame first so that the data can be saved as a CSV
            _spark_session_.createDataFrame(per_identity_data.flatMap(lambda x: x[1])).write.csv(
                path, header=True)

    def print_output(self, per_identity_data) -> None:
        for row in per_identity_data.map(lambda x: json.dumps(x, cls=BlurrJSONEncoder)).collect():
            print(row)



from datetime import datetime, timezone
from typing import Any, Dict, List, Tuple

from dateutil import parser

from blurr.core.schema_loader import SchemaLoader
from blurr.core.store import Store, Key


class MemoryStore(Store):
    """
    In-memory store implementation
    """

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        super().__init__(fully_qualified_name, schema_loader)
        self._cache: Dict[Key, Any] = dict()

    def load(self):
        pass

    def get(self, key: Key) -> Any:
        return self._cache.get(key, None)

    def get_all(self, identity: str = None) -> Dict[Key, Any]:
        return {k: v
                for k, v in self._cache.items()
                if k.identity == identity} if identity else self._cache.copy()

    def get_range(self, start: Key, end: Key = None, count: int = 0) -> List[Tuple[Key, Any]]:
        if end and count:
            raise ValueError('Only one of `end` or `count` can be set')

        if end is not None and end < start:
            start, end = end, start

        if not count:
            # TODO: Come up with a better way of getting this range for instances where key
            # does not have a timestamp.
            items_in_range = []
            for key, item in self._cache.items():
                if start < key < end:
                    items_in_range.append((key, item))
                elif key.timestamp is None:
                    item_ts = parser.parse(item.get('_start_time', datetime.min.isoformat()))
                    item_ts = item_ts if item_ts.tzinfo else item_ts.replace(tzinfo=timezone.utc)
                    if start.timestamp < item_ts < end.timestamp:
                        items_in_range.append((key, item))
                        continue
            return items_in_range
        else:
            filter_condition = (lambda i: i[0] > start) if count > 0 else (lambda i: i[0] < start)

            items = sorted(filter(filter_condition, list(self._cache.items())))

            if abs(count) > len(items):
                count = MemoryStore._sign(count) * len(items)

            if count < 0:
                return items[count:]
            else:
                return items[:count]

    @staticmethod
    def _sign(x: int) -> int:
        return (1, -1)[x < 0]

    def save(self, key: Key, item: Any) -> None:
        self._cache[key] = item

    def delete(self, key: Key) -> None:
        self._cache.pop(key, None)

    def finalize(self) -> None:
        pass



from typing import Dict, Any, List, Tuple, Optional

from blurr.core.errors import InvalidSchemaError
from blurr.core.loader import TypeLoader
from blurr.core.type import Type


class SchemaLoader:
    """
    Provides functionality to operate on the schema using fully qualified names.
    """
    ATTRIBUTE_NAME = 'Name'
    ATTRIBUTE_TYPE = 'Type'
    ITEM_SEPARATOR = '.'

    def __init__(self):
        self._spec = {}
        self._object_cache: Dict[str, 'BaseSchema'] = {}

    def add_schema(self, spec: Dict[str, Any],
                   fully_qualified_parent_name: str = None) -> Optional[str]:
        """
        Add a schema dictionary to the schema loader. The given schema is stored
        against fully_qualified_parent_name + ITEM_SEPARATOR('.') + schema.name.
        :param spec: Schema specification.
        :param fully_qualified_parent_name: Full qualified name of the parent.
        If None is passed then the schema is stored against the schema name.
        :return: The fully qualified name against which the spec is stored.
        None is returned if the given spec is not a dictionary or the spec does not
        contain a 'name' key.
        """
        if not isinstance(spec, dict) or self.ATTRIBUTE_NAME not in spec:
            return None

        name = spec[self.ATTRIBUTE_NAME]
        fully_qualified_name = name if fully_qualified_parent_name is None else self.get_fully_qualified_name(
            fully_qualified_parent_name, name)

        self._spec[fully_qualified_name] = spec
        for key, val in spec.items():
            if isinstance(val, list):
                for item in val:
                    self.add_schema(item, fully_qualified_name)
            self.add_schema(val, fully_qualified_name)

        return spec[self.ATTRIBUTE_NAME]

    # Using forward reference to avoid cyclic dependency.
    def get_schema_object(self, fully_qualified_name: str) -> 'BaseSchema':
        """
        Used to generate a schema object from the given fully_qualified_name.
        :param fully_qualified_name: The fully qualified name of the object needed.
        :return: An initialized schema object
        """

        if fully_qualified_name not in self._object_cache:
            spec = self.get_schema_spec(fully_qualified_name)
            if self.ATTRIBUTE_TYPE not in spec:
                raise InvalidSchemaError('Type not defined in schema')
            self._object_cache[fully_qualified_name] = TypeLoader.load_schema(
                spec[self.ATTRIBUTE_TYPE])(fully_qualified_name, self)

        return self._object_cache[fully_qualified_name]

    def get_nested_schema_object(self, fully_qualified_parent_name: str,
                                 nested_item_name: str) -> 'BaseSchema':
        """
        Used to generate a schema object from the given fully_qualified_parent_name
        and the nested_item_name.
        :param fully_qualified_parent_name: The fully qualified name of the parent.
        :param nested_item_name: The nested item name.
        :return: An initialized schema object of the nested item.
        """
        return self.get_schema_object(
            self.get_fully_qualified_name(fully_qualified_parent_name, nested_item_name))

    @staticmethod
    def get_fully_qualified_name(fully_qualified_parent_name: str, nested_item_name: str) -> str:
        """
        Returns the fully qualified name by combining the fully_qualified_parent_name
        and nested_item_name.
        :param fully_qualified_parent_name: The fully qualified name of the parent.
        :param nested_item_name: The nested item name.
        :return: The fully qualified name of the nested item.
        """
        return fully_qualified_parent_name + SchemaLoader.ITEM_SEPARATOR + nested_item_name

    def get_schema_spec(self, fully_qualified_name: str) -> Dict[str, Any]:
        """
        Used to retrieve the specifications of the schema from the given
        fully_qualified_name of schema.
        :param fully_qualified_name: The fully qualified name of the schema needed.
        :return: Schema dictionary.
        """
        try:
            return self._spec[fully_qualified_name]
        except:
            raise InvalidSchemaError("{} not declared in schema".format(fully_qualified_name))

    def get_schemas_of_type(self, schema_type: Type) -> List[Tuple[str, Dict[str, Any]]]:
        """
        Returns a list of fully qualified names and schema dictionary tuples for
        the schema type provided.
        :param type: Schema type.
        :return: List of fully qualified names and schema dictionary tuples.
        """

        return [(fq_name, schema)
                for fq_name, schema in self._spec.items()
                if Type.is_type_equal(schema.get(self.ATTRIBUTE_TYPE, ''), schema_type)]

    @staticmethod
    def get_transformer_name(fully_qualified_name: str) -> str:
        """
        Returns the DTC transformer name based on the given fully_qualified_name
        of one of the nested child items.
        :param fully_qualified_name: Fully qualified name of the nested child item.
        :return: DTC transformer name.
        """
        return fully_qualified_name.split(SchemaLoader.ITEM_SEPARATOR, 1)[0]



from abc import ABC
from typing import Dict, Type, Any

from blurr.core.base import BaseSchemaCollection, BaseItemCollection, BaseItem
from blurr.core.errors import MissingAttributeError
from blurr.core.evaluation import EvaluationContext
from blurr.core.loader import TypeLoader
from blurr.core.schema_loader import SchemaLoader
from blurr.core.store_key import Key
from blurr.core.type import Type as DTCType


class AggregateSchema(BaseSchemaCollection, ABC):
    """
    Group Schema must inherit from this base.  Data Group schema provides the
    abstraction for managing the 'Fields' in the group.
    """

    # Field Name Definitions
    ATTRIBUTE_STORE = 'Store'
    ATTRIBUTE_FIELDS = 'Fields'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        """
        Initializing the nested field schema that all data groups contain
        :param spec: Schema specifications for the field
        """
        super().__init__(fully_qualified_name, schema_loader, self.ATTRIBUTE_FIELDS)
        self.store = None
        if self.ATTRIBUTE_STORE in self._spec:
            self.store = self._load_store(self._spec[self.ATTRIBUTE_STORE])

    def _load_store(self, store_name: str) -> 'Store':
        """
        Load a store into the Aggregate
        :param store_name: The name of the store
        """
        store_fq_name = self.schema_loader.get_fully_qualified_name(
            self.schema_loader.get_transformer_name(self.fully_qualified_name), store_name)
        return self.schema_loader.get_schema_object(store_fq_name)

    def extend_schema(self, spec: Dict[str, Any]) -> Dict[str, Any]:
        """ Injects the identity field """

        identity_field = {'Name': '_identity', 'Type': DTCType.STRING, 'Value': 'identity'}
        spec[self.ATTRIBUTE_FIELDS].insert(0, identity_field)

        self.schema_loader.add_schema(identity_field, self.fully_qualified_name)

        # If field type is missing, set it to string by default
        for field in spec[self.ATTRIBUTE_FIELDS]:
            if self.ATTRIBUTE_TYPE not in field:
                field[self.ATTRIBUTE_TYPE] = DTCType.STRING

        return super().extend_schema(spec)


class Aggregate(BaseItemCollection, ABC):
    """
    All Data Groups inherit from this base.  Provides an abstraction for 'value' of the encapsulated
    to be called as properties of the data group itself.
    """

    def __init__(self, schema: AggregateSchema, identity: str,
                 evaluation_context: EvaluationContext) -> None:
        """
        Initializes the data group with the inherited context and adds
        its own nested items into the local context for execution
        :param schema: Schema for initializing the data group
        :param evaluation_context: Context dictionary for evaluation
        """
        super().__init__(schema, evaluation_context)
        self._identity = identity

        self._fields: Dict[str, Type[BaseItem]] = {
            name: TypeLoader.load_item(item_schema.type)(item_schema, self._evaluation_context)
            for name, item_schema in self._schema.nested_schema.items()
        }

    @property
    def _nested_items(self) -> Dict[str, Type[BaseItem]]:
        """
        Returns the dictionary of fields the Aggregate contains
        """
        return self._fields

    def finalize(self) -> None:
        """
        Saves the current state of the Aggregate in the store as the final rites
        """
        self.persist()

    def persist(self, timestamp=None) -> None:
        """
        Persists the current data group
        :param timestamp: Optional timestamp to include in the Key construction
        """
        if self._schema.store:
            self._schema.store.save(Key(self._identity, self._name, timestamp), self._snapshot)

    def __getattr__(self, item: str) -> Any:
        """
        Makes the value of the nested items available as properties
        of the collection object.  This is used for retrieving field values
        for dynamic execution.
        :param item: Field requested
        """
        return self.__getitem__(item)

    def __getitem__(self, item: str) -> Any:
        """
        Makes the nested items available though the square bracket notation.
        :raises KeyError: When a requested item is not found in nested items
        """
        if item not in self._nested_items:
            raise MissingAttributeError('{item} not defined in {name}'.format(
                item=item, name=self._name))

        return self._nested_items[item].value



from collections import defaultdict
from datetime import datetime
from typing import Dict, Any

from blurr.core.aggregate_block import BlockAggregate
from blurr.core.base import BaseSchema, BaseItem
from blurr.core.evaluation import Expression, EvaluationContext
from blurr.core.schema_loader import SchemaLoader


class AnchorSchema(BaseSchema):
    """
    Represents the schema for the Anchor specified in a window DTC.
    """
    ATTRIBUTE_CONDITION = 'Condition'
    ATTRIBUTE_MAX = 'Max'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        super().__init__(fully_qualified_name, schema_loader)

        self.condition = Expression(self._spec[self.ATTRIBUTE_CONDITION])
        self.max = self._spec[self.ATTRIBUTE_MAX] if self.ATTRIBUTE_MAX in self._spec else None


class Anchor(BaseItem):
    def __init__(self, schema: AnchorSchema, evaluation_context: EvaluationContext):
        super().__init__(schema, evaluation_context)
        self.condition_met: Dict[datetime, int] = defaultdict(int)
        self.anchor_block = None

    def evaluate_anchor(self, block: BlockAggregate) -> bool:
        self.anchor_block = block
        if self.max_condition_met(block):
            return False

        if self.evaluate():
            return True

        return False

    def add_condition_met(self):
        self.condition_met[self.anchor_block._start_time.date()] += 1

    def evaluate(self) -> bool:
        return self._schema.condition.evaluate(self._evaluation_context)

    def max_condition_met(self, block: BlockAggregate) -> bool:
        if self._schema.max is None:
            return False
        return self.condition_met[block._start_time.date()] >= self._schema.max

    def restore(self, snapshot: Dict[str, Any]) -> BaseItem:
        pass

    def reset(self) -> None:
        self.condition_met: Dict[datetime, int] = defaultdict(int)
        self.anchor_block = None

    @property
    def _snapshot(self):
        pass



import sys
from typing import List

import yaml

from blurr.cli.util import get_yml_files, eprint
from blurr.core.errors import InvalidSchemaError
from blurr.core.syntax.schema_validator import validate


def validate_command(dtc_files: List[str]) -> int:
    all_files_valid = True
    if len(dtc_files) == 0:
        dtc_files = get_yml_files()
    for dtc_file in dtc_files:
        if validate_file(dtc_file) == 1:
            all_files_valid = False

    return 0 if all_files_valid else 1


def validate_file(dtc_file: str) -> int:
    print('Running syntax validation on {}'.format(dtc_file))
    try:
        dtc_dict = yaml.safe_load(open(dtc_file, 'r', encoding='utf-8'))
        validate(dtc_dict)
        print('Document is valid')
        return 0
    except yaml.YAMLError as err:
        eprint('Invalid yaml')
        eprint(str(err))
        return 1
    except InvalidSchemaError as err:
        eprint(str(err))
        return 1
    except:
        exception_value = sys.exc_info()[1]
        logging.error(exception_value)
        eprint('There was an error parsing the document')
        return 1


def get_valid_yml_files(yml_files: List[str]) -> List[str]:
    return [yml_file for yml_file in yml_files if validate_file(yml_file) == 0]



from abc import ABC, abstractmethod
from typing import Dict, Any, Type, TypeVar, Union

from blurr.core.errors import SnapshotError
from blurr.core.evaluation import Expression, EvaluationContext
from blurr.core.schema_loader import SchemaLoader
from blurr.core.store_key import Key


class BaseSchema(ABC):
    """ Encapsulates the common functionality of all schema elements """

    ATTRIBUTE_NAME = 'Name'
    ATTRIBUTE_DESCRIPTION = 'Description'
    ATTRIBUTE_TYPE = 'Type'
    ATTRIBUTE_WHEN = 'When'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        """
        Initializes a schema by providing the path to the schema and the schema loader resource
        :param fully_qualified_name: Fully qualified path to the schema
        :param schema_loader: Schema repository that returns schema spec by fully qualified name
        """
        self.schema_loader = schema_loader
        self.fully_qualified_name = fully_qualified_name
        self._spec: Dict[str, Any] = self.extend_schema(
            self.schema_loader.get_schema_spec(self.fully_qualified_name))

        self.name: str = self._spec[self.ATTRIBUTE_NAME]
        self.type: str = self._spec[self.ATTRIBUTE_TYPE]
        self.when: Expression = Expression(
            self._spec[self.ATTRIBUTE_WHEN]) if self.ATTRIBUTE_WHEN in self._spec else None
        self.description = self._spec.get(self.ATTRIBUTE_DESCRIPTION, None)

    def extend_schema(self, spec: Dict[str, Any]) -> Dict[str, Any]:
        """ Extends the defined schema specifications at runtime with defaults """
        return spec


class BaseSchemaCollection(BaseSchema, ABC):
    """
    Base class for schema that contain nested schema
    """

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader,
                 nested_schema_attribute: str) -> None:
        """
        Initializes a schema collection with nested items
        :param fully_qualified_name: Fully qualified path to the schema
        :param schema_loader: Schema repository that returns schema spec by fully qualified name
        :param nested_schema_attribute: Name of the attribute that contains the nested elements
        """
        super().__init__(fully_qualified_name, schema_loader)

        self._nested_item_attribute = nested_schema_attribute
        # Load nested schema items
        self.nested_schema: Dict[str, Type[BaseSchema]] = {
            schema_spec[self.ATTRIBUTE_NAME]: self.schema_loader.get_nested_schema_object(
                self.fully_qualified_name, schema_spec[self.ATTRIBUTE_NAME])
            for schema_spec in self._spec.get(self._nested_item_attribute, [])
        }


BaseItemType = TypeVar('T', bound='BaseItem')


class BaseItem(ABC):
    """
    Base class for for all leaf items that do not contain sub-items
    """

    def __init__(self, schema: BaseSchema, evaluation_context: EvaluationContext) -> None:
        """
        Initializes an item with the schema and execution context
        :param schema: Schema of the item
        :param evaluation_context: Context dictionary for evaluation
        """
        self._schema = schema
        self._evaluation_context: EvaluationContext = evaluation_context

    @property
    def _needs_evaluation(self) -> bool:
        """
        Returns True when:
            1. Where clause is not specified
            2. Where WHERE clause is specified and it evaluates to True
        Returns false if a where clause is specified and it evaluates to False
        """
        return self._schema.when is None or self._schema.when.evaluate(self._evaluation_context)

    @property
    def _name(self) -> str:
        """
        Returns the name of the base item
        """
        return self._schema.name

    @abstractmethod
    def evaluate(self, *args, **kwargs) -> None:
        """
        Evaluates the current item
        """
        raise NotImplementedError('evaluate() must be implemented')

    @property
    @abstractmethod
    def _snapshot(self):
        """
        Gets a dictionary representation of the current state items in the current hierarchy
        :return: Name, Value map of the current tree
        """
        raise NotImplementedError('snapshot() must be implemented')

    @abstractmethod
    def restore(self, snapshot) -> BaseItemType:
        """
        Restores the state of an item from a snapshot
        """
        raise NotImplementedError('restore() must be implemented')

    @abstractmethod
    def reset(self) -> None:
        """
        Resets the state of the item.
        """
        raise NotImplementedError('reset() must be implemented')


class BaseItemCollection(BaseItem, ABC):
    """
    Base class for items that contain sub-items within them
    """

    def __init__(self, schema: BaseSchemaCollection, evaluation_context: EvaluationContext) -> None:
        """
        Loads nested items to the 'items' collection
        :param schema: Schema that conforms to the item
        :param evaluation_context: Context dictionary for evaluation
        """

        super().__init__(schema, evaluation_context)

    def evaluate(self, *args, **kwargs) -> None:
        """
        Evaluates the current item
        :returns An evaluation result object containing the result, or reasons why
        evaluation failed
        """
        if self._needs_evaluation:
            for _, item in self._nested_items.items():
                item.evaluate()

    @property
    def _snapshot(self) -> Dict[str, Any]:
        """
        Implements snapshot for collections by recursively invoking snapshot of all child items
        """
        try:
            return {name: item._snapshot for name, item in self._nested_items.items()}
        except Exception as e:
            raise SnapshotError('Error while creating snapshot for {}'.format(self._name)) from e

    def restore(self, snapshot: Dict[Union[str, Key], Any]) -> 'BaseItemCollection':
        """
        Restores the state of a collection from a snapshot
        """
        try:

            for name, snap in snapshot.items():
                if isinstance(name, Key):
                    self._nested_items[name.group].restore(snap)
                else:
                    self._nested_items[name].restore(snap)
            return self

        except Exception as e:
            raise SnapshotError('Error while restoring snapshot: {}'.format(self._snapshot)) from e

    def reset(self) -> None:
        for _, item in self._nested_items.items():
            item.reset()

    @abstractmethod
    def finalize(self) -> None:
        """
        Performs the final rites of an item before it decommissioned
        """
        raise NotImplementedError('finalize() must be implemented')

    @property
    @abstractmethod
    def _nested_items(self) -> Dict[str, Type[BaseItem]]:
        """
        Dictionary of the name and item in the collection
        """
        raise NotImplementedError('nested_items() must be implemented')



import os
import re
import traceback
from typing import Dict

from yamale import yamale
from yamale.schema import Data
from yamale.validators import DefaultValidators, Validator
from yamale.validators.constraints import Constraint

from blurr.core.errors import InvalidSchemaError
from blurr.core.type import Type

IDENTITY_VALIDATOR_REGEX = re.compile(r'^_|[^\S]')


class StringExclude(Constraint):
    keywords = {'exclude': list}
    fail = '\'%s\' is a reserved keyword.  Please try another.'

    def _is_valid(self, value):
        return value not in self.exclude

    def _fail(self, value):
        return self.fail % value


class DataType(Validator):
    TAG = 'data_type'

    VALUES = ['integer', 'boolean', 'string', 'datetime', 'float', 'map', 'list', 'set']

    def _is_valid(self, value: str) -> bool:
        return value in self.VALUES

    def get_name(self) -> str:
        return 'DTC Valid Data Type'


class DimensionDataType(Validator):
    TAG = 'dimension_data_type'

    VALUES = ['integer', 'boolean', 'string']

    def _is_valid(self, value: str) -> bool:
        return value in self.VALUES

    def get_name(self) -> str:
        return 'Valid LabelField Type'


class Identifier(Validator):
    TAG = 'identifier'
    constraints = [StringExclude]

    def _is_valid(self, value: str) -> bool:
        return not IDENTITY_VALIDATOR_REGEX.findall(value)

    def get_name(self) -> str:
        return 'Identifier'

    def fail(self, value):
        return '\'%s\' starts with _ or containing whitespace characters.' % value


class Expression(Validator):
    TAG = 'expression'

    ERROR_STRING_INVALID_PYTHON_EXPRESSION = '\'%s\' is an invalid python expression.'
    failure_reason = None

    def _is_valid(self, value: str) -> bool:
        value = str(value)
        try:
            compile(str(value), '<string>', 'eval')
            return True
        except Exception as err:
            traceback.print_exc(limit=0)
            self.failure_reason = self.ERROR_STRING_INVALID_PYTHON_EXPRESSION
        return False

    def get_name(self) -> str:
        return 'Expression'

    def fail(self, value):
        return self.failure_reason % value


VALIDATORS = {
    **DefaultValidators.copy(),
    DataType.TAG: DataType,
    DimensionDataType.TAG: DimensionDataType,
    Identifier.TAG: Identifier,
    Expression.TAG: Expression,
}

PACKAGE_DIR = os.path.dirname(os.path.abspath(__file__))
STREAMING_SCHEMA = yamale.make_schema(
    os.path.join(PACKAGE_DIR, 'dtc_streaming_schema.yml'), validators=VALIDATORS)

WINDOW_SCHEMA = yamale.make_schema(
    os.path.join(PACKAGE_DIR, 'dtc_window_schema.yml'), validators=VALIDATORS)


def _validate_window(dtc_dict: Dict, name: str) -> None:
    try:
        WINDOW_SCHEMA.validate(Data(dtc_dict, name))
    except ValueError as e:
        raise InvalidSchemaError(str(e))


def _validate_streaming(dtc_dict: Dict, name: str) -> None:
    try:
        STREAMING_SCHEMA.validate(Data(dtc_dict, name))
    except ValueError as e:
        raise InvalidSchemaError(str(e))


def is_window_dtc(dtc_dict: Dict) -> bool:
    return Type.is_type_equal(dtc_dict.get('Type', ''), Type.BLURR_TRANSFORM_WINDOW)


def is_streaming_dtc(dtc_dict: Dict) -> bool:
    return Type.is_type_equal(dtc_dict.get('Type', ''), Type.BLURR_TRANSFORM_STREAMING)


def validate(dtc_dict: Dict, name='dtc') -> None:
    if is_window_dtc(dtc_dict):
        _validate_window(dtc_dict, name)
    elif is_streaming_dtc(dtc_dict):
        _validate_streaming(dtc_dict, name)
    else:
        raise InvalidSchemaError('Document has an invalid DTC \'Type\' {}.'.format(
            dtc_dict.get('Type', '')))



class InvalidSchemaError(Exception):
    """
    Indicates an error in the schema specification
    """
    pass


class InvalidExpressionError(Exception):
    """
    Indicates that a python expression specified is either non-compilable, or not allowed
    """
    pass


class ExpressionEvaluationError(Exception):
    """
    Error raised during expression evaluation by the interpreter
    """
    pass


class TypeNotFoundError(Exception):
    """
    Indicates dynamic type loading failure if type is not found type map
    """
    pass


class SnapshotError(Exception):
    """
    Indicates issues with serializing the current state of the object
    """
    pass


class StaleBlockError(Exception):
    """
    Indicates that the event being processed cannot be added to the block rollup that is loaded
    """
    pass


class StreamingSourceNotFoundError(Exception):
    """
    Raised when the raw data for streaming is unavailable in the execution context
    """
    pass


class AnchorBlockNotDefinedError(Exception):
    """
    Raised when anchor block is not defined and a WindowTransformer is evaluated.
    """
    pass


class IdentityError(Exception):
    """
    Raised when there is an error in the identity determination of a record.
    """
    pass


class TimeError(Exception):
    """
    Raised when there is an error in determining the time of the record.
    """
    pass


class PrepareWindowMissingBlocksError(Exception):
    """
    Raised when the window view generated is insufficient as per the window specification.
    """
    pass


class MissingAttributeError(Exception):
    """
    Raised when the name of the item being retrieved does not exist in the nested items.
    """
    pass



from abc import ABC, abstractmethod
from datetime import datetime
from decimal import Decimal
from json import JSONEncoder
from typing import List, Optional, Tuple, Any, Union, Dict, Iterable, Generator

import yaml

from blurr.core import logging
from blurr.core.aggregate_block import BlockAggregate
from blurr.core.errors import PrepareWindowMissingBlocksError
from blurr.core.evaluation import Context
from blurr.core.record import Record
from blurr.core.schema_loader import SchemaLoader
from blurr.core.store import Store
from blurr.core.store_key import Key
from blurr.core.transformer_streaming import StreamingTransformer, StreamingTransformerSchema
from blurr.core.transformer_window import WindowTransformer
from blurr.core.type import Type
from blurr.runner.data_processor import DataProcessor


class BlurrJSONEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            ratio = o.as_integer_ratio()
            if ratio[1] == 1:
                return int(o)
            else:
                return float(o)

        if isinstance(o, Key):
            return str(o)

        return super().default(o)


class Runner(ABC):
    def __init__(self, stream_dtc_file: str, window_dtc_file: Optional[str]):
        self._stream_dtc = yaml.safe_load(open(stream_dtc_file))
        self._window_dtc = None if window_dtc_file is None else yaml.safe_load(
            open(window_dtc_file))

        # TODO: Assume validation will be done separately.
        # This causes a problem when running the code on spark
        # as the validation yml file is inside the archived package
        # and yamale is not able to read that.
        # validate(self._stream_dtc)
        # if self._window_dtc is not None:
        #     validate(self._window_dtc)

    def execute_per_identity_records(self,
                                     identity_records: Tuple[str, List[Tuple[datetime, Record]]]
                                     ) -> List[Union[Tuple[Key, Any], Tuple[str, Any]]]:
        identity, events = identity_records
        schema_loader = SchemaLoader()
        events.sort(key=lambda x: x[0])

        block_data = self._execute_stream_dtc(events, identity, schema_loader)
        window_data = self._execute_window_dtc(identity, schema_loader)

        if self._window_dtc is None:
            return [(k, v) for k, v in block_data.items()]
        else:
            return [(identity, window_data)]

    def get_per_identity_records(self, events: Iterable, data_processor: DataProcessor
                                 ) -> Generator[Tuple[str, Tuple[datetime, Record]], None, None]:
        schema_loader = SchemaLoader()
        stream_dtc_name = schema_loader.add_schema(self._stream_dtc)
        stream_transformer_schema: StreamingTransformerSchema = schema_loader.get_schema_object(
            stream_dtc_name)
        for event in events:
            for record in data_processor.process_data(event):
                try:
                    id = stream_transformer_schema.get_identity(record)
                    time = stream_transformer_schema.get_time(record)
                    yield (id, (time, record))
                except Exception as err:
                    logging.debug('{} in parsing Record.'.format(err))

    def _execute_stream_dtc(self, identity_events: List[Tuple[datetime, Record]], identity: str,
                            schema_loader: SchemaLoader) -> Dict[Key, Any]:
        if self._stream_dtc is None:
            return {}

        stream_dtc_name = schema_loader.add_schema(self._stream_dtc)
        stream_transformer_schema = schema_loader.get_schema_object(stream_dtc_name)

        stream_transformer = StreamingTransformer(stream_transformer_schema, identity)
        for time, event in identity_events:
            stream_transformer.evaluate(event)
        stream_transformer.finalize()

        return self._get_store(schema_loader).get_all(identity)

    def _execute_window_dtc(self, identity: str, schema_loader: SchemaLoader) -> List[Dict]:
        if self._window_dtc is None:
            logging.debug('Window DTC not provided')
            return []

        stream_transformer = StreamingTransformer(
            self._get_streaming_transformer_schema(schema_loader), identity)
        all_data = self._get_store(schema_loader).get_all(identity)
        stream_transformer.restore(all_data)

        exec_context = Context()
        exec_context.add(stream_transformer._schema.name, stream_transformer)

        block_obj = None
        for aggregate in stream_transformer._nested_items.values():
            if not isinstance(aggregate, BlockAggregate):
                continue
            if block_obj is not None:
                raise Exception(('Window operation is supported against Streaming ',
                                 'DTC with only one BlockAggregate'))
            block_obj = aggregate

        if block_obj is None:
            raise Exception('No BlockAggregate found in the Streaming DTC file')

        window_data = []

        window_dtc_name = schema_loader.add_schema(self._window_dtc)
        window_transformer_schema = schema_loader.get_schema_object(window_dtc_name)
        window_transformer = WindowTransformer(window_transformer_schema, identity, exec_context)

        logging.debug('Running Window DTC for identity {}'.format(identity))

        anchors = 0
        blocks = 0
        for key, data in all_data.items():
            if key.group != block_obj._schema.name:
                continue
            try:
                blocks += 1
                if window_transformer.evaluate(block_obj.restore(data)):
                    anchors += 1
                    window_data.append(window_transformer.flattened_snapshot)
            except PrepareWindowMissingBlocksError as err:
                logging.debug('{} with {}'.format(err, key))

        if anchors == 0:
            logging.debug('No anchors found for identity {} out of {} blocks'.format(
                identity, blocks))

        return window_data

    def _get_store(self, schema_loader: SchemaLoader) -> Store:
        store_schemas = schema_loader.get_schemas_of_type(Type.BLURR_STORE_MEMORY)
        if not store_schemas:
            store_schemas = schema_loader.get_schemas_of_type(Type.BLURR_STORE_DYNAMO)
        return schema_loader.get_schema_object(store_schemas[0][0])

    def _get_streaming_transformer_schema(
            self, schema_loader: SchemaLoader) -> StreamingTransformerSchema:
        streaming_transformer_schema = schema_loader.get_schemas_of_type(
            Type.BLURR_TRANSFORM_STREAMING)
        return schema_loader.get_schema_object(streaming_transformer_schema[0][0])

    @abstractmethod
    def execute(self, *args, **kwargs):
        NotImplemented('execute must be implemented')

    @abstractmethod
    def write_output_file(self, *args, **kwargs):
        NotImplemented('execute must be implemented')

    @abstractmethod
    def print_output(self, *args, **kwargs):
        NotImplemented('execute must be implemented')



import os
import sys
from typing import List, Tuple

import yaml

from blurr.core.syntax.schema_validator import is_streaming_dtc, is_window_dtc


def get_yml_files(path: str = '.') -> List[str]:
    return [
        os.path.join(path, file)
        for file in os.listdir(path)
        if (file.endswith('.yml') or file.endswith('.yaml'))
    ]


def get_stream_window_dtc_files(dtc_files: List[str]) -> Tuple[str, str]:
    stream_dtc_file = None
    window_dtc_file = None
    for dtc_file in dtc_files:
        if stream_dtc_file and window_dtc_file:
            break

        dtc_dict = yaml.safe_load(open(dtc_file))
        if not stream_dtc_file and is_streaming_dtc(dtc_dict):
            stream_dtc_file = dtc_file

        if not window_dtc_file and is_window_dtc(dtc_dict):
            window_dtc_file = dtc_file

    return stream_dtc_file, window_dtc_file


def eprint(*args):
    print(*args, file=sys.stderr)



from abc import abstractmethod
from typing import Any, List, Tuple, Dict

from blurr.core.base import BaseSchema
from blurr.core.store_key import Key


class Store(BaseSchema):
    """ Base Store that allows for data to be persisted during / after transformation """

    @abstractmethod
    def get_all(self, identity: str) -> Dict[Key, Any]:
        """
        Gets all the items for an identity
        """
        raise NotImplementedError()

    @abstractmethod
    def get(self, key: Key) -> Any:
        """
        Gets an item by key
        """
        raise NotImplementedError()

    @abstractmethod
    def get_range(self, start: Key, end: Key = None, count: int = 0) -> List[Tuple[Key, Any]]:
        raise NotImplementedError()

    @abstractmethod
    def save(self, key: Key, item: Any) -> None:
        """
        Saves an item to store
        """
        raise NotImplementedError()

    @abstractmethod
    def delete(self, key: Key) -> None:
        """
        Deletes an item from the store by key
        """
        raise NotImplementedError()

    @abstractmethod
    def finalize(self) -> None:
        """
        Finalizes the store by flushing all remaining data to persistence
        """
        raise NotImplementedError()



from typing import Dict, Any, List

from blurr.core.aggregate import Aggregate, AggregateSchema
from blurr.core.evaluation import Expression
from blurr.core.schema_loader import SchemaLoader
from blurr.core.type import Type


class BlockAggregateSchema(AggregateSchema):
    """
    Aggregates that handles the block rollup aggregation
    """

    ATTRIBUTE_SPLIT = 'Split'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        super().__init__(fully_qualified_name, schema_loader)

        # Load type specific attributes
        self.split: Expression = Expression(
            self._spec[self.ATTRIBUTE_SPLIT]) if self.ATTRIBUTE_SPLIT in self._spec else None

    def extend_schema(self, spec: Dict[str, Any]) -> Dict[str, Any]:
        """ Injects the block start and end times """

        # Add new fields to the schema spec
        predefined_field = self._build_time_fields_spec(spec[self.ATTRIBUTE_NAME])
        spec[self.ATTRIBUTE_FIELDS][0:0] = predefined_field

        # Add new field schema to the schema loader
        for field_schema in predefined_field:
            self.schema_loader.add_schema(field_schema, self.fully_qualified_name)

        return super().extend_schema(spec)

    @staticmethod
    def _build_time_fields_spec(name_in_context: str) -> List[Dict[str, Any]]:
        """
        Constructs the spec for predefined fields that are to be included in the master spec prior to schema load
        :param name_in_context: Name of the current object in the context
        :return:
        """
        return [
            {
                'Name': '_start_time',
                'Type': Type.DATETIME,
                'Value': ('time if {aggregate}._start_time is None else time '
                          'if time < {aggregate}._start_time else {aggregate}._start_time'
                          ).format(aggregate=name_in_context)
            },
            {
                'Name': '_end_time',
                'Type': Type.DATETIME,
                'Value': ('time if {aggregate}._end_time is None else time '
                          'if time > {aggregate}._end_time else {aggregate}._end_time'
                          ).format(aggregate=name_in_context)
            },
        ]


class BlockAggregate(Aggregate):
    """
    Manages the aggregates for block based roll-ups of streaming data
    """

    def evaluate(self) -> None:
        """
        Evaluates the current item
        """

        # If a split is imminent, save the current block snapshot with the timestamp
        split_should_be_evaluated = not (self._schema.split is None or self._start_time is None
                                         or self._end_time is None)

        if split_should_be_evaluated and self._schema.split.evaluate(
                self._evaluation_context) is True:
            # Save the current snapshot with the current timestamp
            self.persist(self._start_time)
            # Reset the state of the contents
            self.__init__(self._schema, self._identity, self._evaluation_context)

        super().evaluate()

    def persist(self, timestamp=None) -> None:
        super().persist(timestamp if timestamp else self._start_time)



from datetime import timedelta

from blurr.core.aggregate_block import BlockAggregate, BlockAggregateSchema
from blurr.core.schema_loader import SchemaLoader


class ActivityAggregateSchema(BlockAggregateSchema):
    ATTRIBUTE_SEPARATE_BY_INACTIVE_SECONDS = 'SeparateByInactiveSeconds'

    def __init__(self, fully_qualified_name: str, schema_loader: SchemaLoader) -> None:
        super().__init__(fully_qualified_name, schema_loader)

        self.separation_interval: timedelta = timedelta(
            seconds=int(self._spec[self.ATTRIBUTE_SEPARATE_BY_INACTIVE_SECONDS]))


class ActivityAggregate(BlockAggregate):
    """ Aggregates activity in blocks separated by periods of inactivity"""

    def evaluate(self) -> None:
        time = self._evaluation_context.global_context['time']

        # If the event time is beyond separation threshold, create a new block
        if self._start_time and (time < self._start_time - self._schema.separation_interval
                                 or time > self._end_time + self._schema.separation_interval):
            self.persist(self._start_time)
            self.reset()

        super().evaluate()

    def finalize(self):
        self.persist(self._start_time)



from datetime import datetime, timezone

from dateutil import parser


class Key:
    """
    A record in the store is identified by a key
    """
    PARTITION = '/'

    def __init__(self, identity: str, group: str, timestamp: datetime = None) -> None:
        """
        Initializes a new key for storing data
        :param identity: Primary identity of the record being stored
        :param group: Secondary identity of the record
        :param timestamp: Optional timestamp that can be used for time range queries
        """
        if not identity or identity.isspace():
            raise ValueError('`identity` must be present.')

        if not group or group.isspace():
            raise ValueError('`group` must be present.')

        self.identity = identity
        self.group = group
        self.timestamp = timestamp if not timestamp or timestamp.tzinfo else timestamp.replace(
            tzinfo=timezone.utc)

    @staticmethod
    def parse(key_string: str) -> 'Key':
        """ Parses a flat key string and returns a key """
        parts = key_string.split(Key.PARTITION)
        return Key(parts[0], parts[1], parser.parse(parts[2]) if len(parts) > 2 else None)

    def __str__(self):
        """ Returns the string representation of the key"""
        if self.timestamp:
            return Key.PARTITION.join([self.identity, self.group, self.timestamp.isoformat()])

        return Key.PARTITION.join([self.identity, self.group])

    def __eq__(self, other: 'Key') -> bool:
        return other and (self.identity, self.group,
                          self.timestamp) == (other.identity, other.group, other.timestamp)

    def __lt__(self, other: 'Key') -> bool:
        """
        Does a less than comparison on two keys. A None timestamp is considered
        larger than a timestamp that has been set.
        """
        if (self.identity, self.group) != (other.identity, other.group) or self.timestamp is None:
            return False

        return (other.timestamp is None) or (self.timestamp < other.timestamp)

    def __gt__(self, other: 'Key') -> bool:
        """
        Does a greater than comparison on two keys. A None timestamp is
        considered larger than a timestamp that has been set.
        """
        if (self.identity, self.group) != (other.identity, other.group) or other.timestamp is None:
            return False

        return (self.timestamp is None) or (self.timestamp > other.timestamp)

    def __hash__(self):
        return hash((self.identity, self.group, self.timestamp))
